# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "Lock",
    ]

import os
import fcntl

from fcntl import (
    LOCK_UN,
    LOCK_SH,
    LOCK_EX,
    LOCK_NB,
    )


class Lock:

    def __init__(self, filename):
        self._lock_filename = filename
        self._lock_fd = None
        self._lock_inode = None
        self._lock_kind = LOCK_UN
        self._lock_count = 0

    def __del__(self):
        try:
            os.close(self._lock_fd)
        except:
            pass

    def acquired(self, write=False):
        if write:
            return self._lock_kind == LOCK_EX
        else:
            return self._lock_kind != LOCK_UN

    def acquire(self, write=False, block=True):
        """Acquire the lock for reading or writing.

        :param write: If true, the path is acquired for writing.
        :param block: If false, the call will return immediately on failure.

        This method is reentrant, but not thread-safe.  To use instances
        of this class concurrently across threads, some instructions
        will have to be protected to make it thread-safe.
        """
        if (self._lock_kind == LOCK_EX or
            self._lock_kind == LOCK_SH and not write):
            self._lock_count += 1
            return True

        result = False
        while True:
            if self._lock_fd is None:
                self._lock_fd = os.open(
                    self._lock_filename, os.O_RDWR | os.O_CREAT, 0644)
                flags = fcntl.fcntl(self._lock_fd, fcntl.F_GETFD, 0)
                flags |= fcntl.FD_CLOEXEC
                fcntl.fcntl(self._lock_fd, fcntl.F_SETFD, flags)
                self._lock_inode = os.fstat(self._lock_fd).st_ino
            if write:
                flags = kind = LOCK_EX
            else:
                flags = kind = LOCK_SH
            if not block:
                flags |= LOCK_NB
            try:
                # Locks threads and local system, bad network behavior.
                fcntl.flock(self._lock_fd, flags)
            except IOError:
                # Someone else has the lock and we're not blocking.
                break

            try:
                # Locks local system and network, bad thread behavior.
                fcntl.lockf(self._lock_fd, flags)
            except IOError:
                # Someone else has the lock and we're not blocking.
                fcntl.flock(self._lock_fd, LOCK_UN)
                break

            try:
                lock_inode = os.stat(self._lock_filename).st_ino
                if lock_inode != self._lock_inode:
                    raise OSError
            except OSError:
                # Oops.. someone removed the filename while
                # holding the lock.  Let's continue and try
                # to acquire the new lock.
                os.close(self._lock_fd)
                self._lock_fd = None
                continue

            self._lock_count += 1

            if write or self._lock_kind == LOCK_UN:
                self._lock_kind = kind

            result = True
            break

        return result

    def release(self):
        """Release the lock.

        This method is reentrant, but not thread safe.  To use instances
        of this class concurrently across threads, some instructions
        will have to be protected to make it thread-safe.
        """
        if self._lock_count == 0:
            raise RuntimeError("Lock not acquired")

        self._lock_count -= 1

        if self._lock_count == 0:
            os.close(self._lock_fd)
            os.unlink(self._lock_filename)
            self._lock_fd = None
            self._lock_kind = LOCK_UN
