# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "LaunchpadMixin",
    ]

import shutil

from os import (
    environ,
    path,
    )
from optparse import OptionGroup
from urlparse import urlsplit

# Make launchpadlib happy
try:
    import lazr.restful
except ImportError:
    pass

from lazr.restfulclient import _browser
from lazr.restfulclient.errors import HTTPError

from launchpadlib.launchpad import Launchpad

from lpresults.scripts.application import ApplicationError
from lpresults.scripts.functor import Functor
from lpresults.scripts.retry import retry_except


class LaunchpadMixin:

    # Launchpad defaults
    default_application_name = "Launchpad Results"
    default_service_root = "http://localhost:8000/+api/"
    default_version = "1.0"

    def getLaunchpad(self, *args, **kwargs):
        """Helper to get a launchpad instance based on application options."""
        # Cache is broken with multiple processes, so each child process
        # needs to be given a different cache directory but the same
        # credentials can be shared safely.
        #
        # See: https://bugs.launchpad.net/launchpadlib/+bug/459418
        index = kwargs.pop("index", None)
        if index is not None:
            launchpad_kwargs = self.launchpad_factory._kwargs
            parent_launchpadlib_dir = launchpad_kwargs["launchpadlib_dir"]
            child_launchpadlib_dir = path.join(
                parent_launchpadlib_dir, str(index))
            kwargs.setdefault("launchpadlib_dir", child_launchpadlib_dir)

            # Copy the parent cache directory to the child
            host_name = urlsplit(launchpad_kwargs["service_root"])[1]
            child_cache_dir = path.join(
                child_launchpadlib_dir, host_name, "cache")
            if not path.exists(child_cache_dir):
                parent_cache_dir = path.join(
                    parent_launchpadlib_dir, host_name, "cache")
                shutil.copytree(parent_cache_dir, child_cache_dir)

        try:
            return self.launchpad_factory(*args, **kwargs)
        except AttributeError:
            # In stock httplib2, this exception is raised when failing
            # to connect to the remote host:
            #
            #   AttributeError: 'NoneType' object has no attribute 'makefile'
            #
            # See: http://code.google.com/p/httplib2/issues/detail?id=62
            raise ApplicationError(
                "Failed network connection to service root")

        except HTTPError, error:
            # When launchpadlib fails connecting to the service root,
            # this is an example of exception raised:
            #
            #   HTTPError: HTTP Error 504: Gateway Time-out
            raise ApplicationError(
                "Failed HTTP connection to service root: %s"
                % error.response.reason)

    def addOptions(self, parser):
        """See Application."""
        super(LaunchpadMixin, self).addOptions(parser)

        group = OptionGroup(parser, "Launchpad options")
        group.add_option("--application-name",
            default=self.default_application_name,
            metavar="NAME",
            help="The name of this application,"
                " defaults to: %default")
        group.add_option("--credentials-file",
            metavar="FILE",
            help="If given, the credentials are stored in that file.")
        group.add_option("--launchpadlib-dir",
            metavar="DIR",
            help="The directory for cache and credentials.")
        group.add_option("--service-root",
            default=self.default_service_root,
            metavar="URL",
            help="The URL to the root of the web service,"
                " defaults to: %default")
        group.add_option("--version",
            default=self.default_version,
            help="The version of the web service,"
                " defaults to: %default")
        parser.add_option_group(group)

    def parseOptions(self, options, args):
        """See Application."""
        super(LaunchpadMixin, self).parseOptions(options, args)

        # Define launchpadlib_dir and credentials_file if empty
        launchpadlib_dir = options.launchpadlib_dir
        if launchpadlib_dir is None:
            home_dir = environ["HOME"]
            launchpadlib_dir = path.join(home_dir, ".lpresults")

        credentials_file = options.credentials_file
        if credentials_file is None:
            host_name = urlsplit(options.service_root)[1]
            credentials_file = path.join(
                launchpadlib_dir, host_name, "credentials",
                options.application_name)

        self.launchpad_factory = Functor(
            Launchpad.login_with, options.application_name,
            credentials_file=path.expanduser(credentials_file),
            launchpadlib_dir=path.expanduser(launchpadlib_dir),
            service_root=options.service_root,
            version=options.version)

    def preProcess(self):
        """See Application."""
        super(LaunchpadMixin, self).preProcess()

        # Establish an early Launchpad connection to get the credentials
        self.getLaunchpad()


# An application can sometimes crash with a 5xx error in
# Launchpad.Browser._request(). This monkey patch attempts
# to retry a few times before giving up.
_original_request = _browser.Browser._request


def retry_request(*args, **kwargs):
    try:
        return _original_request(*args, **kwargs)
    except HTTPError, error:
        # 404 errors indicate a data inconsistency: more than one
        # attempt to open the URL is pointless.
        if error.response.status == 404:
            raise

        raise ApplicationError(str(error))

_browser.Browser._request = retry_except(ApplicationError)(
    retry_request)
