# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "create_from_file",
    ]

try:
    import xml.etree.cElementTree as etree
except ImportError:
    import cElementTree as etree

from StringIO import StringIO
from logging import getLogger

from lpresults.registry.formatters import string_to_datetime

from lpresults.xunit.parsers.cpuinfo import CpuinfoParser
from lpresults.xunit.parsers.dmidecode import DmidecodeParser
from lpresults.xunit.parsers.dmisys import DmiSysParser
from lpresults.xunit.parsers.udevadm import UdevadmParser
from lpresults.xunit.protocol import Protocol
from lpresults.xunit.status import Status


class CheckboxResult:

    def __init__(self, test_run_factory, **kwargs):
        self.test_run_factory = test_run_factory
        self.test_run_kwargs = kwargs
        self.test_run = None
        self.devices = []
        self.summary = {}

    def addContext(self, text, command=None):
        if text.strip() == "Command not found.":
            return

        parsers = {
            "cat /proc/cpuinfo": self.parseCpuinfo,
            "dmidecode": DmidecodeParser,
            "grep -r . /sys/class/dmi/id/ 2&gt;/dev/null": DmiSysParser,
            "udevadm info --export-db": UdevadmParser,
            }
        parser = parsers.get(command)
        if parser:
            if not isinstance(text, unicode):
                text = text.decode("latin1")
            stream = StringIO(text)
            p = parser(stream)
            p.run(self)

    def addDevice(self, device):
        if self.test_run.system_unit:
            self.test_run.system_unit.addState(
                bus_name=device.bus, category_name=device.category,
                product_name=device.product, vendor_name=device.vendor,
                product_id=device.product_id, vendor_id=device.vendor_id,
                subproduct_id=device.subproduct_id,
                subvendor_id=device.subvendor_id,
                driver_name=device.driver)

    def addDmiDevice(self, device):
        if device.category == "SYSTEM" and device.vendor != device.product:
            self.test_run.setSystemUnit(
                person=None,
                make=device.vendor,
                model=device.product,
                identifiers=[device.serial])
        elif device.category == "BOARD" and not self.test_run.system_unit:
            self.test_run.setSystemUnit(
                person=None,
                make=device.vendor,
                model=device.product,
                identifiers=[device.serial])

    def addLSBRelease(self, name, value):
        pass

    def addQuestion(self, question):
        answer_to_status = {
            "fail": Status.FAIL,
            "no": Status.FAIL,
            "pass": Status.PASS,
            "skip": Status.UNTESTED,
            "uninitiated": Status.UNINITIATED,
            "unresolved": Status.UNRESOLVED,
            "unsupported": Status.UNSUPPORTED,
            "untested": Status.UNTESTED,
            "yes": Status.PASS,
            }

        status = answer_to_status[question["answer"]["value"]]
        if self.test_run:
            self.test_run.addTestResult(
                name=question["name"],
                output=question["comment"],
                status=status,
                **self.test_run_kwargs)

    def addSummary(self, name, value):
        self.summary[name] = value
        # Build the test run once the distribution and distroseries are known
        if (name == "distribution" and "distroseries" in self.summary) \
           or (name == "distroseries" and "distribution" in self.summary):
            self.test_run = self.test_run_factory(
                project=unicode(self.summary["distribution"]),
                series=unicode(self.summary["distroseries"]),
                **self.test_run_kwargs)

    def setProcessor(self, processor):
        pass

    def parseCpuinfo(self, stream):
        architecture_to_machine = {
            "amd64": "x86_64",
            }
        architecture = self.summary["architecture"]
        machine = architecture_to_machine.get(architecture, architecture)
        return CpuinfoParser(stream, machine)


class CheckboxProtocol(Protocol):

    def __init__(self, file):
        self.file = file

        self.logger = getLogger()

    def _getClient(self, node):
        """Return a dictionary with the name and version of the client."""
        return {
            "name": node.get("name"),
            "version": node.get("version"),
            }

    def _getProperty(self, node):
        """Return the (name, value) of a property."""
        return (node.get("name"), self._getValueAsType(node))

    def _getProperties(self, node):
        """Return a dictionary of properties."""
        properties = {}
        for child in node.getchildren():
            assert child.tag == "property", \
                "Unexpected tag <%s>, expected <property>" % child.tag
            name, value = self._getProperty(child)
            properties[name] = value

        return properties

    def _getValueAsType(self, node):
        """Return value of a node as the type attribute."""
        type_ = node.get("type")
        if type_ in ("bool",):
            value = node.text.strip()
            assert value in ("True", "False",), \
                "Unexpected boolean value '%s' in <%s>" % (value, node.tag)
            return value == "True"
        elif type_ in ("str",):
            return node.text.strip()
        elif type_ in ("int", "long",):
            return int(node.text.strip())
        elif type_ in ("float",):
            return float(node.text.strip())
        elif type_ in ("list",):
            return list(self._getValueAsType(child)
                for child in node.getchildren())
        elif type_ in ("dict",):
            return dict((child.get("name"), self._getValueAsType(child))
                for child in node.getchildren())
        else:
            raise AssertionError(
                "Unexpected type '%s' in <%s>" % (type_, node.tag))

    def _getValueAsBoolean(self, node):
        """Return the value of the attribute "value" as a boolean."""
        value = node.attrib["value"]
        assert value in ("True", "False",), \
            "Unexpected boolean value '%s' in tag <%s>" % (value, node.tag)
        return value == "True"

    def _getValueAsDatetime(self, node):
        """Return the value of the attribute "value" as a datetime."""
        string = node.attrib["value"]
        return string_to_datetime(string)

    def _getValueAsString(self, node):
        """Return the value of the attribute "value"."""
        return node.attrib["value"]

    def parseContext(self, result, node):
        """Parse the <context> part of a submission."""
        for child in node.getchildren():
            assert child.tag == "info", \
                "Unexpected tag <%s>, expected <info>" % child.tag
            result.addContext(child.text, child.get("command"))

    def parseHardware(self, result, node):
        parsers = {
            "dmi": DmidecodeParser,
            "udev": UdevadmParser,
            }

        for child in node.getchildren():
            parser = parsers.get(child.tag)
            if parser:
                text = child.text
                if not isinstance(text, unicode):
                    text = text.decode("latin1")
                stream = StringIO(text)
                p = parser(stream)
                p.run(result)

    def parseLSBRelease(self, result, node):
        """Parse the <lsb_release> part of a submission."""
        for name, value in self._getProperties(node).items():
            result.addLSBRelease(name, value)

    def parseQuestions(self, result, node):
        """Parse the <questions> part of a submission."""
        for child in node.getchildren():
            assert child.tag == "question", \
                "Unexpected tag <%s>, expected <question>" % child.tag
            question = {
                "name": child.get("name"),
                "targets": [],
                }
            plugin = child.get("plugin", None)
            if plugin is not None:
                question["plugin"] = plugin

            answer_choices = []
            for sub_node in child.getchildren():
                sub_tag = sub_node.tag
                if sub_tag == "answer":
                    question["answer"] = answer = {}
                    answer["type"] = sub_node.get("type")
                    if answer["type"] == "multiple_choice":
                        question["answer_choices"] = answer_choices
                    unit = sub_node.get("unit", None)
                    if unit is not None:
                        answer["unit"] = unit
                    answer["value"] = sub_node.text.strip()

                elif sub_tag == "answer_choices":
                    for value_node in sub_node.getchildren():
                        answer_choices.append(
                            self._getValueAsType(value_node))

                elif sub_tag == "target":
                    # The Relax NG schema ensures that the attribute
                    # id exists and that it is an integer
                    target = {"id": int(sub_node.get("id"))}
                    target["drivers"] = drivers = []
                    for driver_node in sub_node.getchildren():
                        drivers.append(driver_node.text.strip())
                    question["targets"].append(target)

                elif sub_tag in ("comment", "command",):
                    data = sub_node.text
                    if data is not None:
                        question[sub_tag] = data.strip()

                else:
                    raise AssertionError(
                        "Unexpected tag <%s> in <question>" % sub_tag)

            result.addQuestion(question)

    def parseSoftware(self, result, node):
        """Parse the <software> section of a submission."""
        parsers = {
            "lsbrelease": self.parseLSBRelease,
            }

        for child in node.getchildren():
            parser = parsers.get(child.tag)
            if parser:
                parser(result, child)
            else:
                self.logger.debug(
                    "Unsupported tag <%s> in <software>" % child.tag)

    def parseSummary(self, result, node):
        """Parse the <summary> section of a submission."""
        parsers = {
            "architecture": self._getValueAsString,
            "client": self._getClient,
            "contactable": self._getValueAsBoolean,
            "date_created": self._getValueAsDatetime,
            "distribution": self._getValueAsString,
            "distroseries": self._getValueAsString,
            "kernel-release": self._getValueAsString,
            "live_cd": self._getValueAsBoolean,
            "private": self._getValueAsBoolean,
            "system_id": self._getValueAsString,
            }

        for child in node.getchildren():
            parser = parsers.get(child.tag)
            if parser:
                value = parser(child)
                result.addSummary(child.tag, value)
            else:
                self.logger.debug(
                    "Unsupported tag <%s> in <summary>" % child.tag)

    def parseRoot(self, result, node):
        """Parse the <system> root of a submission."""
        parsers = {
            "context": self.parseContext,
            "hardware": self.parseHardware,
            "questions": self.parseQuestions,
            "software": self.parseSoftware,
            "summary": self.parseSummary,
            }

        # Iterate over the root children, "summary" first
        children = sorted(
            node.getchildren(), key=lambda k: k.tag != "summary")
        for child in children:
            parser = parsers.get(child.tag)
            if parser:
                parser(result, child)
            else:
                self.logger.debug(
                    "Unsupported tag <%s> in <system>" % child.tag)

    def run(self, test_run_factory, **kwargs):
        parser = etree.XMLParser()

        try:
            tree = etree.parse(self.file, parser=parser)
        except SyntaxError, error_value:
            self.logger.error(error_value)
            return None

        root = tree.getroot()
        if root.tag != "system":
            self.logger.error(
                "Unexpected tag <%s> at root, expected <system>" % root.tag)
            return None

        result = CheckboxResult(test_run_factory, **kwargs)
        try:
            self.parseRoot(result, root)
        except AssertionError, error:
            self.logger.error(error)
            return None

        return result


def create_from_file(file):
    header = """<?xml version="1.0" ?>"""
    buffer = file.read(len(header))
    file.seek(0)

    protocol = None
    if buffer == header:
        protocol = CheckboxProtocol(file)

    return protocol
