/* Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
 * GNU Affero General Public License version 3 (see the file LICENSE).
 *
 * The Action module provides an abstraction layer to run client
 * requests in a deferred context.
 *
 * @module lpresults.command
 * @submodule action
 */
YUI.add('lpresults.command.action', function(Y) {

    var namespace = Y.namespace('lpresults.command');

    var HOST = 'host';

    /**
     * The Action class provides the ability to run a function and get
     * a Deferred on which additional callbacks can be defined.
     *
     * @class Action
     * @extends Base
     * @constructor
     */
    var Action = function(config) {
        Action.superclass.constructor.apply(this, arguments);
    };

    Action.NAME = 'action';
    Action.ATTRS = {
        /**
         * The client function to use. If not defined, one will be created
         * during initialization.
         *
         * @attribute client
         */
        client: {
            value: null
        }
    };

    Y.extend(Action, Y.Base, {
        /**
         * Run the client with the given parameters. Return a Deferred firing
         * with the result of the action, or failing with an error if the
         * action failed.
         *
         * @method run
         * @public
         */
       run: function(parameters) {
            var deferred = new Y.lpresults.command.Deferred();

            var client = this.get('client');
            if (client !== null) {
                try {
                    result = client(parameters);
                    this.runSuccess.apply(this, [deferred, result]);
                } catch (error) {
                    this.runFailure.apply(this, [deferred, error]);
                }
            }

            return deferred;
        },

        /**
         * Handle a failed run action.
         *
         * @method runFailure
         * @protected
         */
        runFailure: function(deferred, result) {
            deferred.errback(result);
        },

        /**
         * Handle a successful run action.
         *
         * @method runSuccess
         * @protected
         */
        runSuccess: function(deferred, result) {
            deferred.callback(result);
        }
    });

    /*
     * Ensure that Action is available within the command namespace.
     */
    namespace.Action = Action;

    /**
     * Plugin for binding an action to a Button widget.
     *
     * @class ActionPlugin
     * @extends Plugin
     * @constructor
     */
    var ActionPlugin = function(config) {
        ActionPlugin.superclass.constructor.apply(this, arguments);
    };

    ActionPlugin.NAME = 'action_plugin';
    ActionPlugin.NS = 'action';
    ActionPlugin.ATTRS = {
        /**
         * The action to execute on click.
         *
         * @attribute action
         */
        action: {
            value: null
        },

        /**
         * Validate the form before running the action.
         *
         * @attribute validate
         * @type Boolean
         */
        validate: {
            value: false
        }
    };

    Y.extend(ActionPlugin, Y.Plugin.Base, {
        /**
         * Constructor logic.
         *
         * @method initializer
         * @protected
         */
        initializer: function(config) {
            this.afterHostMethod('bindUI', this._bindUI);
            if (this.get(HOST).get('rendered')) {
                this._bindUI();
            }
        },

        /**
         * Bind click events to action.
         */
        _bindUI: function() {
            var host = this.get(HOST);
            host.set('onclick', {fn: this._onClick, scope: this});
        },

        /**
         * Run the action using the form input.
         */
        _onClick: function() {
            var host = this.get(HOST);

            if (this.get('validate')) {
                if (!host.get('parent')._runValidation()) {
                    return;
                }
            }

            this.get('action').run();
        }
    });

    /*
     * Ensure that ActionPlugin is available within the command namespace.
     */
    namespace.ActionPlugin = ActionPlugin;

}, '0.1', {'requires': ['base', 'plugin', 'lpresults.command.deferred']});
